# Author: Alex DelFranco
# Advisor: Rafa Martin Domenech
# Institution: Center for Astrophysics | Harvard & Smithsonian
# Date: 27 September 2022

from .main_menu import main_menu

class puma(main_menu):
    pass
